@@ -1,6 +1,6 @@
 """
 --- ___init___.py ---
 """
-__all__ = ["auxiliar_funcs", "jai"]
+__all__ = ["functions", "jai"]
 
 from .jai import Jai
\ No newline at end of file
