@@ -4,12 +4,12 @@ import pandas as pd
 import numpy as np
 import requests
 import time
-import re
 
 from io import BytesIO
 from .processing import process_similar, process_resolution
 from .functions.utils_funcs import data2json, pbar_steps
 from .functions.classes import PossibleDtypes, Mode
+from .functions import exceptions
 from fnmatch import fnmatch
 import matplotlib.pyplot as plt
 from pandas.api.types import is_integer_dtype
@@ -285,13 +285,53 @@ class Jai:
         """
         # find a way to process this
         # what errors to raise, etc.
-        print(f"\n\nSTATUS: {response.status_code}\n\n")
-        raise ValueError(f"Something went wrong.\n{response.content}")
+        message = f"Something went wrong.\n\nSTATUS: {response.status_code}\n"
+        try:
+            detail = response.json()['detail']
+            if "Error: " in detail:
+                error, msg = detail.split(": ", 1)
+                message += msg
+                try:
+                    raise eval(error)(message)
+                except NameError:
+                    try:
+                        raise eval("exceptions." + error)(message)
+                    except NameError:
+                        raise ValueError(message)
+            else:
+                message += detail
+                raise ValueError(message)
+        except TypeError:
+            message += str(response.text)
+            raise ValueError(message)
+
+    def filters(self, name):
+        """
+        Gets the valid values of filters.
+
+        Args
+        ----
+        name : str
+            String with the name of a database in your JAI environment.
+
+        Return
+        ------
+        response : list of strings
+            List of valid filter values.
+        """
+        response = requests.get(self.url + f"/filters/{name}",
+                                headers=self.header)
+
+        if response.status_code == 200:
+            return response.json()
+        else:
+            return self.assert_status_code(response)
 
     def similar(self,
                 name: str,
                 data,
                 top_k: int = 5,
+                filters=None,
                 batch_size: int = 16384):
         """
         Query a database in search for the `top_k` most similar entries for each
@@ -346,7 +386,10 @@ class Jai:
                     _batch = data[i:i + batch_size].tolist()
                 else:
                     _batch = data[i:i + batch_size].tolist()
-                res = self._similar_id(name, _batch, top_k=top_k)
+                res = self._similar_id(name,
+                                       _batch,
+                                       top_k=top_k,
+                                       filters=filters)
             else:
                 if isinstance(data, (pd.Series, pd.DataFrame)):
                     _batch = data.iloc[i:i + batch_size]
@@ -354,11 +397,16 @@ class Jai:
                     _batch = data[i:i + batch_size]
                 res = self._similar_json(name,
                                          data2json(_batch, dtype=dtype),
-                                         top_k=top_k)
+                                         top_k=top_k,
+                                         filters=filters)
             results.extend(res["similarity"])
         return results
 
-    def _similar_id(self, name: str, id_item: list, top_k: int = 5):
+    def _similar_id(self,
+                    name: str,
+                    id_item: list,
+                    top_k: int = 5,
+                    filters=None):
         """
         Creates a list of dicts, with the index and distance of the k items most similars given an id.
         This is a protected method.
@@ -384,8 +432,11 @@ class Jai:
             raise TypeError(
                 f"id_item param must be int or list, {type(id_item)} found.")
 
+        filtering = "" if filters is None else "".join(
+            ["&filters=" + s for s in filters])
+        url = self.url + f"/similar/id/{name}?top_k={top_k}" + filtering
         response = requests.put(
-            self.url + f"/similar/id/{name}?top_k={top_k}",
+            url,
             headers=self.header,
             data=json.dumps(id_item),
         )
@@ -395,7 +446,11 @@ class Jai:
         else:
             return self.assert_status_code(response)
 
-    def _similar_json(self, name: str, data_json, top_k: int = 5):
+    def _similar_json(self,
+                      name: str,
+                      data_json,
+                      top_k: int = 5,
+                      filters=None):
         """
         Creates a list of dicts, with the index and distance of the k items most similars given a JSON data entry.
         This is a protected method
@@ -417,7 +472,9 @@ class Jai:
         response : dict
             Dictionary with the index and distance of `the k most similar items`.
         """
-        url = self.url + f"/similar/data/{name}?top_k={top_k}"
+        filtering = "" if filters is None else "".join(
+            ["&filters=" + s for s in filters])
+        url = self.url + f"/similar/data/{name}?top_k={top_k}" + filtering
 
         response = requests.put(url, headers=self.header, data=data_json)
         if response.status_code == 200:
@@ -572,6 +629,7 @@ class Jai:
               db_type: str,
               batch_size: int = 16384,
               frequency_seconds: int = 1,
+              filter_name: str = None,
               verbose: int = 1,
               **kwargs):
         """
@@ -632,6 +690,7 @@ class Jai:
         insert_responses = self._insert_data(data=data,
                                              name=name,
                                              batch_size=batch_size,
+                                             filter_name=filter_name,
                                              db_type=db_type)
 
         # check if we inserted everything we were supposed to
@@ -639,8 +698,21 @@ class Jai:
 
         # train model
         setup_response = self._setup_database(name, db_type, **kwargs)
+        if "kwargs" in setup_response:
+            print("\nRecognized setup args:")
+            for key, value in setup_response["kwargs"].items():
+                kwarg_input = kwargs.get(key, None)
+                if isinstance(kwarg_input, dict):
+                    value = json.loads(value)
+                    intersection = kwarg_input.keys() & value.keys()
+                    value = {k: value[k] for k in intersection}
+                print(f"- {key}: {value}")
 
         if frequency_seconds >= 1:
+            if db_type in ["Supervised", "SelfSupervised"]:
+                print(
+                    f"\nTraining might finish early due to early stopping criteria."
+                )
             self.wait_setup(name=name, frequency_seconds=frequency_seconds)
 
         if db_type in [
@@ -672,6 +744,7 @@ class Jai:
                  data,
                  batch_size: int = 16384,
                  frequency_seconds: int = 1,
+                 filter_name: str = None,
                  predict: bool = False):
         """
         Insert raw data and extract their latent representation.
@@ -714,6 +787,7 @@ class Jai:
                                              name=name,
                                              batch_size=batch_size,
                                              db_type=db_type,
+                                             filter_name=filter_name,
                                              predict=predict)
 
         # check if we inserted everything we were supposed to
@@ -764,7 +838,13 @@ class Jai:
         else:
             return self.assert_status_code(response)
 
-    def _insert_data(self, data, name, db_type, batch_size, predict=False):
+    def _insert_data(self,
+                     data,
+                     name,
+                     db_type,
+                     batch_size,
+                     filter_name: str = None,
+                     predict=False):
         """
         Insert raw data for training. This is a protected method.
 
@@ -790,11 +870,15 @@ class Jai:
         for i, b in enumerate(
                 trange(0, len(data), batch_size, desc="Insert Data")):
             _batch = data.iloc[b:b + batch_size]
-            insert_responses[i] = self._insert_json(
-                name, data2json(_batch, dtype=db_type, predict=predict))
+            data_json = data2json(_batch,
+                                  dtype=db_type,
+                                  filter_name=filter_name,
+                                  predict=predict)
+            insert_responses[i] = self._insert_json(name, data_json,
+                                                    filter_name)
         return insert_responses
 
-    def _insert_json(self, name: str, df_json):
+    def _insert_json(self, name: str, df_json, filter_name: str = None):
         """
         Insert data in JSON format. This is a protected method.
 
@@ -810,9 +894,9 @@ class Jai:
         response : dict
             Dictionary with the API response.
         """
-        response = requests.post(self.url + f"/data/{name}",
-                                 headers=self.header,
-                                 data=df_json)
+        filtering = "" if filter_name is None else f"?filter_name={filter_name}"
+        url = self.url + f"/data/{name}" + filtering
+        response = requests.post(url, headers=self.header, data=df_json)
         if response.status_code == 200:
             return response.json()
         else:
@@ -852,34 +936,9 @@ class Jai:
             raise ValueError(f"missing arguments {missing}")
 
         body = {}
-        flag = True
         for key in possible:
             val = kwargs.get(key, None)
             if val is not None:
-                if flag:
-                    print("Recognized setup args:")
-                    flag = False
-
-                if key == "hyperparams":
-                    if "patience" in val and int(val["patience"]) < 1:
-                        val["patience"] = 10  # default patience value for our purposes
-                        print(
-                            f"'patience' value must be greater than or equal to 1, but got {val['patience']} instead. Setting it to 10 (default)"
-                        )
-
-                    if "min_delta" in val and float(val["min_delta"]) < 0:
-                        val["min_delta"] = 1e-5  # default min_delta value for our purposes
-                        print(
-                            f"'min_delta' value must be greater than or equal to 0, but got {val['min_delta']} instead. Setting it to 1e-5 (default)"
-                        )
-
-                    if "max_epochs" in val and int(val["max_epochs"]) < 1:
-                        val["max_epochs"] = 500  # default max_epochs value for our purposes
-                        print(
-                            f"'max_epochs' value must be greater than or equal to 1, but got {val['max_epochs']} instead. Setting it to 500 (default)"
-                        )
-
-                print(f"{key}: {val}")
                 body[key] = val
 
         body["db_type"] = db_type
@@ -948,7 +1007,6 @@ class Jai:
 
         if response.status_code == 200:
             result = response.json()
-            result.pop("Auto lr finder", None)
 
             if 'Model Training' in result.keys():
                 plots = result['Model Training']
@@ -960,6 +1018,7 @@ class Jai:
                 plt.xlabel("epoch")
                 plt.show()
 
+            print("\nSetup Report:")
             print(result['Model Evaluation']
                   ) if 'Model Evaluation' in result.keys() else None
             print()
@@ -1108,7 +1167,7 @@ class Jai:
                 "'fields' method is only available to dtype SelSupervised and Supervised."
             )
 
-        response = requests.get(self.url + f"/table/fields/{name}",
+        response = requests.get(self.url + f"/fields/{name}",
                                 headers=self.header)
         if response.status_code == 200:
             return response.json()
@@ -1149,7 +1208,8 @@ class Jai:
         try:
             with tqdm(total=max_steps,
                       desc="JAI is working",
-                      bar_format='{l_bar}{bar}|{n_fmt}/{total_fmt}') as pbar:
+                      bar_format='{l_bar}{bar}|{n_fmt}/{total_fmt} [{elapsed}]'
+                      ) as pbar:
                 while status['Status'] != 'Task ended successfully.':
                     if status['Status'] == 'Something went wrong.':
                         raise BaseException(status['Description'])
@@ -1157,9 +1217,6 @@ class Jai:
                         # create a second progress bar to track
                         # training progress
                         _, max_iterations = get_numbers(status)
-                        print(
-                            f"Training might not take {max_iterations} steps due to early stopping criteria."
-                        )
                         with tqdm(total=max_iterations,
                                   desc=f"[{name}] Training",
                                   leave=False) as iteration_bar:
@@ -1178,7 +1235,6 @@ class Jai:
                             # full when early stopping is reached -- peace of mind
                             iteration_bar.update(max_iterations -
                                                  iteration_bar.n)
-                            print("\nDone training.")
 
                     if (step == starts_at) and (aux == 0):
                         pbar.update(starts_at)
@@ -1208,6 +1264,38 @@ class Jai:
         self._delete_status(name)
         return status
 
+    def delete_ids(self, name, ids):
+        """
+        Delete the specified ids from database.
+
+        Args
+        ----
+        name : str
+            String with the name of a database in your JAI environment.
+
+        ids : list
+            List of ids to be removed from database.
+
+        Return
+        -------
+        response : dict
+            Dictionary with the API response.
+
+        Example
+        ----------
+        >>> name = 'chosen_name'
+        >>> j = Jai(AUTH_KEY)
+        >>> j.delete_raw_data(name=name)
+        'All raw data from database 'chosen_name' was deleted!'
+        """
+        response = requests.delete(self.url + f"/entity/{name}",
+                                   headers=self.header,
+                                   data=json.dumps(ids))
+        if response.status_code == 200:
+            return response.json()
+        else:
+            return self.assert_status_code(response)
+
     def delete_raw_data(self, name: str):
         """
         Delete raw data. It is good practice to do this after training a model.
